#ifndef __MESH_QUADRATURE_H
#define __MESH_QUADRATURE_H

/*
* This code isdistrubuted under GPLv3
* 
* Code Author: Miroslav Stoyanov, Oct 2011
* 
* Copyright (C) 2011  Miroslav Stoyanov
* 
* This program is free software: you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation, either version 3 of the License, or
* (at your option) any later version.
* 
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
* 
* Since the GNU General Public License is longer than this entire code, 
* a copy of it can be obtained separately at <http://www.gnu.org/licenses/>
* 
*/

#include "sparse_matrix.h"

struct GeometryTriangle{
        double x[3], y[3];
        double xi2x, xi2y, et2x, et2y, det;
        // x, y - the triangle vertexes
        // xi2x, xi2y, et2x, et2y - form a matrix for the transformation between the triangle and a canonical triangle
        // det - the determinant of the matrix, also twice the area of the triangle
};

struct GeometryQuadrature{
        int num_points;
        double *w, *x, *y, *xi, *et;
        // w - the quadrature weights
        // x,y - the points associated with the real element
        // xi,eta - the points associated with the canonical element
};

struct GeometryElement{
	struct GeometryTriangle T;
	int iu[3], igu[3]; // local and global indexes for the nodes
	int nbrs[3]; // neighbouring elements, or -1 is this is a boundary element
	double u[3]; // the solution for this element
};

struct GeometryMesh{
	int num_elem; // number of elements
	int num_global_u; // number of global nodes
	int num_u; // number of degrees of freedom
	struct GeometryElement *elem; // elements
};

void GeometryTriangleSet( double x0, double y0, double x1, double y1, double x2, double y2, struct GeometryTriangle *T );

void GeometryQuadratureAdd6PointsTriangle( struct GeometryTriangle T, struct GeometryQuadrature *Q );
// forms a 6 point quadrature over the triangle, the quadrature is exact for polynomials of degree up to and including 4

void GeometryQuadratureAdd3PointsTriangle( struct GeometryTriangle T, struct GeometryQuadrature *Q );
// this quadrature is exact only for quadratic polynomials, it is easier to computer, but it is not used anywhere in this code

void GeometryQuadratureFree( struct GeometryQuadrature *Q );
// releases the memory used by the quadrature

void GeometryMeshFormUnit( struct GeometryMesh *mesh, double dx, double dy );
// forms a mesh over the Unit square using element size dx and dy for x and y direction

void GeometryMeshFree( struct GeometryMesh *mesh );
// release all memory allcoated to the mesh

/////////////////////////////////////////////////////////////
// ----- Function to tie together the matrix and mesh ---- //
/////////////////////////////////////////////////////////////

void GeometryFormLHS( struct GeometryMesh mesh, struct Pmat *L );
// forms the left hand side matrix

void GeometryFromRHS( struct GeometryMesh mesh, double *R, double (*f)(double,double), double (*boundary)(double,double) );
// forms the right hand side vector, 
// boundary should return the values of the solution at different points on the bounary, this function will only be called for pointa on the boundary

void GeometryLoadSolution( struct GeometryMesh *mesh, double *u, double (*boundary)(double,double) );
// sets the u values in all the elements from either u (if they are interior) or boundary (if they are on the boundary)

double GeometryComputerErrorL2( struct GeometryMesh mesh, double (*u_exact)(double,double) );
// compute the L2 norm of the error (assumes that the solution has been loaded with GeometryLoadSolution)

void GeometrySaveSolution( struct GeometryMesh mesh, char *filename );
// saves the solution in ascii format that can be read and visualized by Matlab

#endif
