#ifndef __SPARSE_MATRIX_H
#define __SPARSE_MATRIX_H

/*
* This code isdistrubuted under GPLv3
* 
* Code Author: Miroslav Stoyanov, Oct 2011
* 
* Copyright (C) 2011  Miroslav Stoyanov
* 
* This program is free software: you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation, either version 3 of the License, or
* (at your option) any later version.
* 
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
* 
* Since the GNU General Public License is longer than this entire code, 
* a copy of it can be obtained separately at <http://www.gnu.org/licenses/>
* 
*/


#include <stdlib.h>
#include <stdio.h>
#include <math.h>

// umfpack is part of SuiteSparse
#include "umfpack.h"

struct Pnode{ // a node for the sparse pattern
	int indx;
	struct Pnode *next;
};

struct Pattern{ // sparse pattern
	int num_rows;
	int *count;
	struct Pnode **sp;
};

struct Pmat{ // sparse matrix structure (row packed)
	int num_rows;
	int *pntr; // pointers to the rows
	int *indx; // indexes of the rows
	double *vals; // associated values
};

struct FactorizedMILU{
	int num_rows;
	int *pntr; // pointers to the rows
	int *indx; // indexes of the rows
	double *vals; // associated values
	int *indxD; // the diagonal entries
	int *indxT; // the corresponding transposed entrie
	int *pntrC; // pointers for the columns
	int *indxC; // indexes for the columns
};

struct FactorizedUMFPACK{
    char format;
    int num_rows;
    int *pntr;
    int *indx;
    double *vals;
    void * Numeric;
};


// this helps generate a sparsity pattern for the matrix
void pattern_init( struct Pattern *p, int n ); // initialize a new pattern with n rows
void pattern_add( struct Pattern *p, int r, int indx ); // add entrie to the pattern
void pattern_del( struct Pattern *p ); // delete the pattern (release memory)


// functions for the parse matrix
void pmat_create( struct Pmat *m, struct Pattern p ); // create a matrix with the pattern
void pmat_delete( struct Pmat *m ); // release all memory used by m
void pmat_addval( struct Pmat *m, int r, int c, double val ); // increment entrie at (r,c) by val
void pmat_repval( struct Pmat *m, int r, int c, double val ); // replace the entrie at (r,c) with val
// WARNING: if the the index isn't properly set, addval and repval enter into an infinite loop, use smpack_pmat_test_index if uncertain
int pmat_test_index( struct Pmat *m, int r, int c ); // returns 0 if the pattenrn contains r,c and 1 otherwise

void pmat_fwrite_ascii( struct Pmat m, const char *filename ); // writes m into an ascii file
void pmat_fread_ascii( struct Pmat *m, const char *filename ); // read matrix form an ascii faile

void pmat_act( struct Pmat m, double *x, double *r, double alpha );

void pmat_free( struct Pmat *m ); // release all memory that has been alocated

// solvers
// this creates a MILU factorization of a sparse matrix
void MILU_analyze( struct Pmat m, struct FactorizedMILU *ilu );
void MILU_factorize( struct Pmat m, struct FactorizedMILU *ilu );
void MILU_solve( struct FactorizedMILU *ilu, double *x );
void MILU_free( struct FactorizedMILU *ilu );

int GMRES( void (*funcA)(double*,double*,void*), double *x, double *b, void *ud, int size, int max_inner, int max_outer, double tol );
// solves Ax=b, i.e. funcA( x, r, ud ) -> r == b
// x contains the initial guess
// size is the problem size, max_inner and max_outer are the maximum number of inner and outer iterations
// tol is the desired tolerance
// the method returns 0 if convergence is reached and 1 if the final error exceeds tol

// functions to interface with SuiteSparse
void pmat_UMFPACK_factorize( struct Pmat m, struct FactorizedUMFPACK *umf );
void pmat_UMFPACK_solve( struct FactorizedUMFPACK umf, double *b, double *x );
void pmat_UMFPACK_free( struct FactorizedUMFPACK *umf );

void pmat_UMFPACK( struct Pmat m, double *b, double *x );


///////////////////////////
// --- BLAS ROUTINES --- //
///////////////////////////
// this should link to external blas routines
// SuiteSparse requires a full implementation of Blas so we might as well include external (hopefully optimized) blas
//
// blas was originally written in Fortran so all the parameters are called by reference (as opposed to value)
// even the integer constant
// the code below is a wrapper around blas to make a more logical (from C point of view) interface

double ddot_( const int *N, const double *a, const int *inca, const double *b, const int *incb );
double dnrm2_( const int *N, const double *a, const int *inca );
void dcopy_(  const int *N, const double *a, const int *inca, const double *b, const int *incb );
void daxpy_( const int *N, const double *alpha, const double *a, const int *inca, const double *b, const int *incb );
void dscal_( const int *N, const double *alpha, const double *a, const int *inca );


void dcopy( int N, double *a, int lda, double *b, int ldb ); // gives b = a
void daxpy( int N, double alpha, double *a, int lda, double *b, int ldb ); // gives b = b + alpha * a
void dscal( int N, double alpha, double *a, int lda ); // scales a by alpha
double dnrm2( int N, double *a, int lda ); // returns the L-2 norm of a (in R^N finite dimensions)
double ddot( int N, double *a, int lda, double *b, int ldb ); // return the dot product of a and b

double ddoti( int N, double *a, int *indx, double *x ); // this a sparse blas extension routine, computes the inner product of a sparse and dense vectors
// note that the standard implementation assumes 1-indexing (as in Fortran), hence in C we pass &(x[1]) as opposed to &(x[0]) (&(x[0]) = x)
// ddoti is often times not included in the blas implementations, hence I keep the function as is (i.e. no optimized)

#endif
